@@ -1,6 +1,6 @@
 /*
  *   stunnel       TLS offloading and load-balancing proxy
- *   Copyright (C) 1998-2022 Michal Trojnara <Michal.Trojnara@stunnel.org>
+ *   Copyright (C) 1998-2023 Michal Trojnara <Michal.Trojnara@stunnel.org>
  *
  *   This program is free software; you can redistribute it and/or modify it
  *   under the terms of the GNU General Public License as published by the
@@ -37,63 +37,49 @@
 
 #include "prototypes.h"
 
-#ifdef USE_WIN32
-
-DISK_FILE *file_open(char *name, FILE_MODE mode) {
+DISK_FILE *file_fdopen(int fd, FILE_MODE file_mode) {
     DISK_FILE *df;
-    LPTSTR tname;
-    HANDLE fh;
-    DWORD desired_access, creation_disposition;
+    FILE *f;
+    char *mode;
 
-    /* open file */
-    switch(mode) {
-    case FILE_MODE_READ:
-        desired_access=GENERIC_READ;
-        creation_disposition=OPEN_EXISTING;
+    switch(fd) {
+    case 0:
+        f=stdin;
         break;
-    case FILE_MODE_APPEND:
-            /* reportedly more compatible than FILE_APPEND_DATA */
-        desired_access=GENERIC_WRITE;
-        creation_disposition=OPEN_ALWAYS; /* keep the data */
+    case 1:
+        f=stdout;
         break;
-    case FILE_MODE_OVERWRITE:
-        desired_access=GENERIC_WRITE;
-        creation_disposition=CREATE_ALWAYS; /* remove the data */
+    case 2:
+        f=stderr;
         break;
-    default: /* invalid mode */
-        return NULL;
+    default:
+        switch(file_mode) {
+        case FILE_MODE_READ:
+            mode="r";
+            break;
+        case FILE_MODE_APPEND:
+            mode="a";
+            break;
+        case FILE_MODE_OVERWRITE:
+            mode="w";
+            break;
+        default: /* invalid file_mode */
+            return NULL;
+        }
+        f=fdopen(fd, mode);
     }
-    tname=str2tstr(name);
-    fh=CreateFile(tname, desired_access, FILE_SHARE_READ, NULL,
-        creation_disposition, FILE_ATTRIBUTE_NORMAL, (HANDLE)NULL);
-    str_free(tname); /* str_free() overwrites GetLastError() value */
-    if(fh==INVALID_HANDLE_VALUE)
+    if(!f)
         return NULL;
-    if(mode==FILE_MODE_APPEND) /* workaround for FILE_APPEND_DATA */
-        SetFilePointer(fh, 0, NULL, FILE_END);
-
-    /* setup df structure */
     df=str_alloc(sizeof(DISK_FILE));
-    df->fh=fh;
+    df->f=f;
     return df;
 }
 
-#else /* USE_WIN32 */
-
-DISK_FILE *file_fdopen(int fd) {
-    DISK_FILE *df;
-
-    df=str_alloc(sizeof(DISK_FILE));
-    df->fd=fd;
-    return df;
-}
-
-DISK_FILE *file_open(char *name, FILE_MODE mode) {
-    DISK_FILE *df;
+DISK_FILE *file_open(char *name, FILE_MODE file_mode) {
     int fd, flags;
 
     /* open file */
-    switch(mode) {
+    switch(file_mode) {
     case FILE_MODE_READ:
         flags=O_RDONLY;
         break;
@@ -103,66 +89,46 @@ DISK_FILE *file_open(char *name, FILE_MODE mode) {
     case FILE_MODE_OVERWRITE:
         flags=O_CREAT|O_WRONLY|O_TRUNC;
         break;
-    default: /* invalid mode */
+    default: /* invalid file_mode */
         return NULL;
     }
-#ifdef O_NONBLOCK
-    flags|=O_NONBLOCK;
-#elif defined O_NDELAY
-    flags|=O_NDELAY;
-#endif
 #ifdef O_CLOEXEC
     flags|=O_CLOEXEC;
 #endif /* O_CLOEXEC */
+    /* don't fopen() directly to prevent O_CLOEXEC race condition */
+#ifdef USE_WIN32
+    fd=_open(name, flags, _S_IREAD|_S_IWRITE);
+#else /* USE_WIN32 */
     fd=open(name, flags, 0640);
-    if(fd==INVALID_SOCKET)
-        return NULL;
-
-    /* setup df structure */
-    df=str_alloc(sizeof(DISK_FILE));
-    df->fd=fd;
-    return df;
-}
-
 #endif /* USE_WIN32 */
+    if(fd<0)
+        return NULL;
+    return file_fdopen(fd, file_mode);
+}
 
 void file_close(DISK_FILE *df) {
     if(!df) /* nothing to do */
         return;
-#ifdef USE_WIN32
-    CloseHandle(df->fh);
-#else /* USE_WIN32 */
-    if(df->fd>2) /* never close stdin/stdout/stder */
-        close(df->fd);
-#endif /* USE_WIN32 */
+    if(fileno(df->f)>2) /* never close stdin/stdout/stder */
+        fclose(df->f);
     str_free(df);
 }
 
 ssize_t file_getline(DISK_FILE *df, char *line, int len) {
-    /* this version is really slow, but performance is not important here */
-    /* (no buffering is implemented) */
     ssize_t i;
-#ifdef USE_WIN32
-    DWORD num;
-#else /* USE_WIN32 */
-    ssize_t num;
-#endif /* USE_WIN32 */
+    int c;
 
     if(!df) /* not opened */
         return -1;
 
     for(i=0; i<len-1; i++) {
-#ifdef USE_WIN32
-        ReadFile(df->fh, line+i, 1, &num, NULL);
-#else /* USE_WIN32 */
-        num=read(df->fd, line+i, 1);
-#endif /* USE_WIN32 */
-        if(num!=1) { /* EOF */
-            if(i) /* any previously retrieved data */
-                break;
-            else
+        c=getc(df->f);
+        if(c==EOF) {
+            if(!i) /* no previously retrieved data */
                 return -1;
+            break; /* MSDOS-style last file line */
         }
+        line[i]=(char)c;
         if(line[i]=='\n') /* LF */
             break;
         if(line[i]=='\r') /* CR */
@@ -172,30 +138,33 @@ ssize_t file_getline(DISK_FILE *df, char *line, int len) {
     return i;
 }
 
-ssize_t file_putline(DISK_FILE *df, char *line) {
+ssize_t file_putline_nonewline(DISK_FILE *df, char *line) {
+    /* used for fatal_debug() -> no str.c functions are allowed */
+    FILE *f;
+    int num;
+
+    f=df ? df->f : stderr; /* no file -> write to stderr */
+    num=fputs(line, f); /* automatically converts LF->CRLF on Windows */
+    return (ssize_t)num;
+}
+
+ssize_t file_putline_newline(DISK_FILE *df, char *line) {
     char *buff;
     size_t len;
-#ifdef USE_WIN32
-    DWORD num;
-#else /* USE_WIN32 */
     ssize_t num;
-#endif /* USE_WIN32 */
 
     len=strlen(line);
-    buff=str_alloc(len+2); /* +2 for CR+LF */
+    buff=str_alloc(len+3); /* +2 for LF+NUL */
     strcpy(buff, line);
-#ifdef USE_WIN32
-    buff[len++]='\r'; /* CR */
-#endif /* USE_WIN32 */
     buff[len++]='\n'; /* LF */
-#ifdef USE_WIN32
-    WriteFile(df->fh, buff, (DWORD)len, &num, NULL);
-#else /* USE_WIN32 */
-    /* no file -> write to stderr */
-    num=write(df ? df->fd : 2, buff, len);
-#endif /* USE_WIN32 */
+    buff[len]='\0'; /* NUL */
+    num=file_putline_nonewline(df, buff);
     str_free(buff);
-    return (ssize_t)num;
+    return num;
+}
+
+int file_flush(DISK_FILE *df) {
+    return fflush(df ? df->f : stderr); /* no file -> flush stderr */
 }
 
 int file_permissions(const char *file_name) {
@@ -212,6 +181,7 @@ int file_permissions(const char *file_name) {
             "Insecure file permissions on %s", file_name);
 #else
     (void)file_name; /* squash the unused parameter warning */
+    /* not (yet) implemented */
 #endif
     return 0;
 }
