@@ -1,6 +1,6 @@
 /*
  *   stunnel       TLS offloading and load-balancing proxy
- *   Copyright (C) 1998-2021 Michal Trojnara <Michal.Trojnara@stunnel.org>
+ *   Copyright (C) 1998-2023 Michal Trojnara <Michal.Trojnara@stunnel.org>
  *
  *   This program is free software; you can redistribute it and/or modify it
  *   under the terms of the GNU General Public License as published by the
@@ -35,14 +35,14 @@
  *   forward this exception.
  */
 
-#include "common.h"
 #include "prototypes.h"
 
 NOEXPORT void log_queue(SERVICE_OPTIONS *, int, char *, char *, char *);
 NOEXPORT void log_raw(SERVICE_OPTIONS *, int, char *, char *, char *);
 NOEXPORT void safestring(char *);
 
-static DISK_FILE *outfile=NULL;
+DISK_FILE *outfile=NULL;
+
 static struct LIST { /* single-linked list of log lines */
     struct LIST *next;
     SERVICE_OPTIONS *opt;
@@ -140,7 +140,22 @@ void log_close(int sink) {
 
 void s_log(int level, const char *format, ...) {
     va_list ap;
-    char *text, *stamp, *id;
+
+    va_start(ap, format);
+    s_vlog(level, format, ap);
+    va_end(ap);
+}
+
+#ifdef __GNUC__
+#if __GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6)
+#pragma GCC diagnostic push
+#endif /* __GNUC__>=4.6 */
+#pragma GCC diagnostic ignored "-Wformat-nonliteral"
+#endif /* __GNUC__ */
+void s_vlog(int level, const char *format, va_list ap) {
+    va_list aq;
+    char stamp[72], id[72], *text;
+    int len;
 #ifdef USE_WIN32
     DWORD libc_error;
 #else
@@ -173,15 +188,25 @@ void s_log(int level, const char *format, ...) {
 #else
         timeptr=localtime(&gmt);
 #endif
-        stamp=str_printf("%04d.%02d.%02d %02d:%02d:%02d",
+        snprintf(stamp, sizeof stamp, "%04d.%02d.%02d %02d:%02d:%02d",
             timeptr->tm_year+1900, timeptr->tm_mon+1, timeptr->tm_mday,
             timeptr->tm_hour, timeptr->tm_min, timeptr->tm_sec);
-        id=str_printf("LOG%d[%s]", level, tls_data->id);
+        snprintf(id, sizeof id, "LOG%d[%s]", level, tls_data->id);
 
         /* format the text to be logged */
-        va_start(ap, format);
-        text=str_vprintf(format, ap);
-        va_end(ap);
+        va_copy(aq, ap);
+        len=vsnprintf(NULL, 0, format, ap);
+        if(len>1024)
+            len=1024;
+#ifdef USE_WIN32
+        text=_alloca((size_t)len+1);
+#else
+        text=alloca((size_t)len+1);
+#endif
+        len=vsnprintf(text, (size_t)len+1, format, aq);
+        va_end(aq);
+        while(len>0 && text[len-1]=='\n')
+            text[--len]='\0'; /* strip trailing newlines */
         safestring(text);
 
         /* either log or queue for logging */
@@ -196,6 +221,11 @@ void s_log(int level, const char *format, ...) {
     set_last_error(libc_error);
     set_last_socket_error(socket_error);
 }
+#ifdef __GNUC__
+#if __GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6)
+#pragma GCC diagnostic pop
+#endif /* __GNUC__>=4.6 */
+#endif /* __GNUC__ */
 
 NOEXPORT void log_queue(SERVICE_OPTIONS *opt,
         int level, char *stamp, char *id, char *text) {
@@ -206,12 +236,9 @@ NOEXPORT void log_queue(SERVICE_OPTIONS *opt,
     tmp->next=NULL;
     tmp->opt=opt;
     tmp->level=level;
-    tmp->stamp=stamp;
-    str_detach(tmp->stamp);
-    tmp->id=id;
-    str_detach(tmp->id);
-    tmp->text=text;
-    str_detach(tmp->text);
+    tmp->stamp=str_dup_detached(stamp);
+    tmp->id=str_dup_detached(id);
+    tmp->text=str_dup_detached(text);
 
     /* append the new element to the list */
     CRYPTO_THREAD_write_lock(stunnel_locks[LOCK_LOG_BUFFER]);
@@ -237,6 +264,9 @@ void log_flush(LOG_MODE new_mode) {
             struct LIST *tmp=head;
             head=head->next;
             log_raw(tmp->opt, tmp->level, tmp->stamp, tmp->id, tmp->text);
+            str_free(tmp->stamp);
+            str_free(tmp->id);
+            str_free(tmp->text);
             str_free(tmp);
         }
         head=tail=NULL;
@@ -249,6 +279,7 @@ void log_flush(LOG_MODE new_mode) {
 NOEXPORT void log_raw(SERVICE_OPTIONS *opt,
         int level, char *stamp, char *id, char *text) {
     char *line;
+    size_t size;
 
     /* NOTE: opt->log_level may have changed since s_log().
      * It is important to use the new value and not the old one. */
@@ -256,33 +287,44 @@ NOEXPORT void log_raw(SERVICE_OPTIONS *opt,
     /* build the line and log it to syslog/file if configured */
     switch(log_mode) {
     case LOG_MODE_CONFIGURED:
-        line=str_printf("%s %s: %s", stamp, id, text);
+        size=strlen(stamp)+strlen(id)+strlen(text)+4;
+#ifdef USE_WIN32
+        line=_alloca(size);
+#else
+        line=alloca(size);
+#endif
+        snprintf(line, size, "%s %s: %s", stamp, id, text);
         if(level<=opt->log_level) {
 #if !defined(USE_WIN32) && !defined(__vms)
             if(global_options.option.log_syslog)
                 syslog(level, "%s: %s", id, text);
 #endif /* USE_WIN32, __vms */
-            if(outfile)
-                file_putline(outfile, line);
+            if(outfile) {
+                file_putline_newline(outfile, line);
+#ifndef USE_OS_THREADS
+                file_flush(outfile);
+#endif /* !USE_OS_THREADS */
+            }
         }
         break;
     case LOG_MODE_ERROR:
         /* don't log the id or the time stamp */
+        size=strlen(text)+5;
+#ifdef USE_WIN32
+        line=_alloca(size);
+#else
+        line=alloca(size);
+#endif
         if(level>=0 && level<=7) /* just in case */
-            line=str_printf("[%c] %s", "***!:.  "[level], text);
+            snprintf(line, size, "[%c] %s", "***!:.  "[level], text);
         else
-            line=str_printf("[?] %s", text);
+            snprintf(line, size, "[?] %s", text);
         break;
     default: /* LOG_MODE_INFO */
         /* don't log the level, the id or the time stamp */
-        line=str_dup(text);
+        line=text;
     }
 
-    /* free the memory */
-    str_free(stamp);
-    str_free(id);
-    str_free(text);
-
     /* log the line to the UI (GUI, stderr, etc.) */
     if(log_mode==LOG_MODE_ERROR ||
             (log_mode==LOG_MODE_INFO && level<LOG_DEBUG) ||
@@ -294,8 +336,6 @@ NOEXPORT void log_raw(SERVICE_OPTIONS *opt,
 #endif
             )
         ui_new_log(line);
-
-    str_free(line);
 }
 
 #ifdef __GNUC__
@@ -307,7 +347,14 @@ NOEXPORT void log_raw(SERVICE_OPTIONS *opt,
 #endif /* __GNUC__ */
 char *log_id(CLI *c) {
     const char table[62]=
-        "0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ";
+        {'0', '1', '2', '3', '4', '5', '6', '7',
+         '8', '9', 'a', 'b', 'c', 'd', 'e', 'f',
+         'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n',
+         'o', 'p', 'q', 'r', 's', 't', 'u', 'v',
+         'w', 'x', 'y', 'z', 'A', 'B', 'C', 'D',
+         'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L',
+         'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T',
+         'U', 'V', 'W', 'X', 'Y', 'Z'};
     unsigned char rnd[22];
     char *uniq;
     size_t i;
@@ -357,10 +404,9 @@ char *log_id(CLI *c) {
 #pragma GCC diagnostic ignored "-Wunused-result"
 #endif /* __GNUC__>=4.6 */
 #endif /* __GNUC__ */
-void fatal_debug(char *txt, const char *file, int line) {
+void fatal_debug(const char *txt, const char *file, int line) {
     char msg[80];
 #ifdef USE_WIN32
-    DWORD num;
 #ifdef UNICODE
     TCHAR tmsg[80];
 #endif
@@ -370,13 +416,8 @@ void fatal_debug(char *txt, const char *file, int line) {
         "INTERNAL ERROR: %s at %s, line %d\n", txt, file, line);
 
     if(outfile) {
-#ifdef USE_WIN32
-        WriteFile(outfile->fh, msg, (DWORD)strlen(msg), &num, NULL);
-#else /* USE_WIN32 */
-        /* no file -> write to stderr */
-        /* no meaningful way here to handle the result */
-        write(outfile ? outfile->fd : 2, msg, strlen(msg));
-#endif /* USE_WIN32 */
+        file_putline_nonewline(outfile, msg);
+        file_flush(outfile);
     }
 
 #ifndef USE_WIN32
@@ -546,7 +587,9 @@ NOEXPORT void safestring(char *c) {
 /* provide hex string corresponding to the input string
  * will be NULL terminated */
 void bin2hexstring(const unsigned char *in_data, size_t in_size, char *out_data, size_t out_size) {
-    const char hex[16]="0123456789ABCDEF";
+    const char hex[16]=
+        {'0', '1', '2', '3', '4', '5', '6', '7',
+         '8', '9' ,'A', 'B', 'C', 'D', 'E', 'F'};
     size_t i;
 
     for(i=0; i<in_size && 2*i+2<out_size; ++i) {
