@@ -1,6 +1,6 @@
 /*
  *   stunnel       TLS offloading and load-balancing proxy
- *   Copyright (C) 1998-2022 Michal Trojnara <Michal.Trojnara@stunnel.org>
+ *   Copyright (C) 1998-2023 Michal Trojnara <Michal.Trojnara@stunnel.org>
  *
  *   This program is free software; you can redistribute it and/or modify it
  *   under the terms of the GNU General Public License as published by the
@@ -93,9 +93,9 @@ NOEXPORT LEAK_ENTRY leak_hash_table[LEAK_TABLE_SIZE],
     *leak_results[LEAK_TABLE_SIZE];
 NOEXPORT int leak_result_num=0;
 
-#if OPENSSL_VERSION_NUMBER >= 0x10100000L
+#if OPENSSL_VERSION_NUMBER >= 0x10101000L
 DEFINE_STACK_OF(LEAK_ENTRY)
-#endif /* OpenSSL version >= 1.1.0 */
+#endif /* OpenSSL version >= 1.1.1 */
 
 #ifdef USE_WIN32
 NOEXPORT LPTSTR str_vtprintf(LPCTSTR, va_list);
@@ -107,7 +107,9 @@ NOEXPORT ALLOC_LIST *get_alloc_list_ptr(void *, const char *, int);
 NOEXPORT void str_leak_debug(const ALLOC_LIST *, int);
 
 NOEXPORT LEAK_ENTRY *leak_search(const ALLOC_LIST *);
+#if OPENSSL_VERSION_NUMBER >= 0x10101000L
 NOEXPORT int leak_cmp(const LEAK_ENTRY *const *, const LEAK_ENTRY *const *);
+#endif /* OpenSSL version >= 1.1.1 */
 NOEXPORT void leak_report(void);
 NOEXPORT long leak_threshold(void);
 
@@ -117,14 +119,6 @@ NOEXPORT volatile uint64_t canary_initialized=CANARY_UNINTIALIZED;
 
 /**************************************** string manipulation functions */
 
-#ifndef va_copy
-#ifdef __va_copy
-#define va_copy(dst, src) __va_copy((dst), (src))
-#else /* __va_copy */
-#define va_copy(dst, src) memcpy(&(dst), &(src), sizeof(va_list))
-#endif /* __va_copy */
-#endif /* va_copy */
-
 char *str_dup_debug(const char *str, const char *file, int line) {
     char *retval;
 
@@ -563,9 +557,9 @@ NOEXPORT LEAK_ENTRY *leak_search(const ALLOC_LIST *alloc_list) {
 void leak_table_utilization() {
     int i, utilization=0;
     int64_t grand_total=0;
-#if OPENSSL_VERSION_NUMBER >= 0x10100000L
+#if OPENSSL_VERSION_NUMBER >= 0x10101000L
     STACK_OF(LEAK_ENTRY) *stats;
-#endif /* OpenSSL version >= 1.1.0 */
+#endif /* OpenSSL version >= 1.1.1 */
 
     /* leak_hash_table[] is only filled at the DEBUG logging level */
     if(service_options.log_level<LOG_DEBUG)
@@ -580,7 +574,7 @@ void leak_table_utilization() {
     s_log(LOG_DEBUG, "Leak detection table utilization: %d/%d (%05.2f%%)",
         utilization, LEAK_TABLE_SIZE, 100.0*utilization/LEAK_TABLE_SIZE);
 
-#if OPENSSL_VERSION_NUMBER >= 0x10100000L
+#if OPENSSL_VERSION_NUMBER >= 0x10101000L
     /* log up to 5 most frequently used heap allocations */
     stats=sk_LEAK_ENTRY_new_reserve(leak_cmp, utilization);
     for(i=0; i<LEAK_TABLE_SIZE; ++i)
@@ -594,9 +588,10 @@ void leak_table_utilization() {
             entry->alloc_file, entry->alloc_line);
     }
     sk_LEAK_ENTRY_free(stats);
-#endif /* OpenSSL version >= 1.1.0 */
+#endif /* OpenSSL version >= 1.1.1 */
 }
 
+#if OPENSSL_VERSION_NUMBER >= 0x10101000L
 NOEXPORT int leak_cmp(const LEAK_ENTRY *const *a, const LEAK_ENTRY *const *b) {
     int64_t d = (*a)->total - (*b)->total;
     if(d>0)
@@ -605,6 +600,7 @@ NOEXPORT int leak_cmp(const LEAK_ENTRY *const *a, const LEAK_ENTRY *const *b) {
         return -1;
     return 0;
 }
+#endif /* OpenSSL version >= 1.1.1 */
 
 /* report identified leaks */
 NOEXPORT void leak_report() {
